from scidb.core import Database, Bucket, DataSet, Data
from typing import Union, List
from pathlib import Path


def get_by_path(bucket: Bucket, path: List[str]) -> Union[Bucket, DataSet, None]:
    target = bucket
    for segment in path:
        if target is None:
            return None
        else:
            target = target.get_data_set(segment, include_deleted=True)
    return target


def import_file(file: Union[str, Path],
                name: str,
                bucket: Bucket,
                path: List[str],
                metadata: Union[None, dict] = None,
                properties: Union[None, dict] = None,
                allow_overwrite: bool = False,
                confirm: bool = True,
                feedback: bool = False) -> Union[None, Data]:
    if confirm and not feedback:
        return None
    if isinstance(file, str):
        file = Path(file)
    if not file.exists():
        raise FileNotFoundError
    target = get_by_path(bucket, path)
    if target is None:
        raise FileNotFoundError
    if not allow_overwrite and target.get_data(name):
        return None
    if target.get_data(name):
        target.delete_data(name, confirm, feedback)
    data = target.add_data(name)
    data.import_file(file, allow_overwrite, confirm, feedback)
    data.set_metadata(metadata)
    data.set_properties(properties)
    return data


def import_dir(directory: Union[str, Path],
               name: str,
               bucket: Bucket,
               path: List[str],
               skip_dir: bool = False,
               metadata: Union[None, dict] = None,
               properties: Union[None, dict] = None,
               allow_overwrite: bool = False,
               confirm: bool = True,
               feedback: bool = False) -> Union[None, DataSet]:
    if confirm and not feedback:
        return None
    if isinstance(directory, str):
        directory = Path(directory)
    if not directory.exists():
        raise FileNotFoundError
    if not directory.is_dir():
        raise NotADirectoryError
    target = get_by_path(bucket, path)
    if target is None:
        raise FileNotFoundError
    if not target.get_data_set(name, include_deleted=True):
        target.add_data_set(name)
    data_set = target.get_data_set(name, include_deleted=True)
    inner_path = path.copy()
    inner_path.append(name)
    for child in directory.glob('*'):
        if child.is_file():
            if not allow_overwrite and data_set.get_data(child.name):
                continue
            data_set.add_data(child.name).import_file(child, allow_overwrite, confirm, feedback)
        elif child.is_dir() and not skip_dir:
            import_dir(
                directory=child,
                name=child.name,
                bucket=bucket,
                path=inner_path,
                allow_overwrite=allow_overwrite,
                confirm=confirm,
                feedback=feedback
            )
    data_set.set_metadata(metadata)
    data_set.set_properties(properties)
    return data_set


def import_tree(root: Union[str, Path],
                bucket_name: str,
                db: Database,
                metadata: Union[None, dict] = None,
                properties: Union[None, dict] = None,
                allow_overwrite: bool = False,
                confirm: bool = True,
                feedback: bool = False) -> Union[None, Bucket]:
    if confirm and not feedback:
        return None
    if isinstance(root, str):
        root = Path(root)
    if not root.exists():
        raise FileNotFoundError
    if not root.is_dir():
        raise NotADirectoryError
    if not db.get_bucket(bucket_name, include_deleted=True):
        db.add_bucket(bucket_name)
    bucket = db.get_bucket(bucket_name, include_deleted=True)
    inner_path = []
    for child in root.glob('*'):
        if child.is_file():
            print(f'WARING: File {child.name} cannot be imported into bucket {bucket_name}.')
        elif child.is_dir():
            import_dir(
                directory=child,
                name=child.name,
                bucket=bucket,
                path=inner_path,
                allow_overwrite=allow_overwrite,
                confirm=confirm,
                feedback=feedback
            )
    bucket.set_metadata(metadata)
    bucket.set_properties(properties)
    return bucket



from scidb.core import Bucket, DataSet
from typing import Callable, Union


def iter_data_set(bucket_or_data_set: Union[Bucket, DataSet],
                  func: Callable,
                  include_deleted: bool = False,
                  depth: Union[None, int] = None,
                  **kwargs):
    if depth is not None and depth <= 0:
        return
    search_list = bucket_or_data_set.all_data_sets if include_deleted else bucket_or_data_set.data_sets
    for child in search_list:
        func(child, **kwargs)
        iter_data_set(child, func, include_deleted, depth=None if depth is None else depth - 1, **kwargs)


def iter_data(bucket_or_data_set: Union[Bucket, DataSet],
              func: Callable,
              include_deleted: bool = False,
              depth: Union[None, int] = None,
              **kwargs):
    if depth is not None and depth <= 0:
        return
    search_list = bucket_or_data_set.all_data_sets if include_deleted else bucket_or_data_set.data_sets
    for child in search_list:
        if isinstance(child, DataSet):
            for data in child.data:
                func(data, **kwargs)
        iter_data(child, func, include_deleted, depth=None if depth is None else depth - 1, **kwargs)



from scidb.core import Database, Bucket, DataSet, Data
from typing import Union, Tuple, List
from pathlib import Path


NODE_TYPE = {
    'database': Database,
    'bucket': Bucket,
    'dataset': DataSet,
    'data': Data
}


class NodeWrapper:
    def __init__(self, node: Union[None, Database, Bucket, DataSet, Data]):
        self.__node__ = node

    @property
    def node(self):
        return self.__node__

    def __rshift__(self, child: Union[str, Tuple[type, str]]) -> 'NodeWrapper':
        node_identifier = ''
        node_type = None
        if isinstance(child, str):
            node_identifier = child
        elif isinstance(child, tuple):
            node_type, node_identifier = child
        if self.__node__ is None:
            raise LookupError
        elif isinstance(self.__node__, Database):
            if node_type in [None, Bucket]:
                return NodeWrapper(self.__node__.get_bucket(node_identifier, include_deleted=True))
            else:
                raise TypeError
        elif isinstance(self.__node__, Bucket):
            if node_type in [None, DataSet]:
                return NodeWrapper(self.__node__.get_data_set(node_identifier, include_deleted=True))
            else:
                raise TypeError
        elif isinstance(self.__node__, DataSet):
            if node_type in [None, DataSet, Data]:
                data_set = self.__node__.get_data_set(node_identifier, include_deleted=True)
                data = self.__node__.get_data(node_identifier)
                if data_set or node_type is DataSet:
                    return NodeWrapper(data_set)
                elif data or node_type is Data:
                    return NodeWrapper(data)
                else:
                    return NodeWrapper(None)
            else:
                raise TypeError
        elif isinstance(self.__node__, Data):
            raise LookupError
        else:
            raise TypeError

    def __lshift__(self, child: Union[str, Tuple[type, str]]) -> 'NodeWrapper':
        node_identifier = ''
        node_type = None
        if isinstance(child, str):
            node_identifier = child
        elif isinstance(child, tuple):
            node_type, node_identifier = child
        if self.__node__ is None:
            raise LookupError
        elif isinstance(self.__node__, Database):
            if node_type in [None, Bucket]:
                child_node = self.__node__.get_bucket(node_identifier, include_deleted=True)
                if child_node is not None:
                    return NodeWrapper(child_node)
                else:
                    return NodeWrapper(self.__node__.add_bucket(node_identifier))
            else:
                raise TypeError
        elif isinstance(self.__node__, Bucket):
            if node_type in [None, DataSet]:
                child_node = self.__node__.get_data_set(node_identifier, include_deleted=True)
                if child_node is not None:
                    return NodeWrapper(child_node)
                else:
                    return NodeWrapper(self.__node__.add_data_set(node_identifier))
            else:
                raise TypeError
        elif isinstance(self.__node__, DataSet):
            if node_type in [None, DataSet]:
                child_node = self.__node__.get_data_set(node_identifier, include_deleted=True)
                if child_node is not None:
                    return NodeWrapper(child_node)
                else:
                    return NodeWrapper(self.__node__.add_data_set(node_identifier))
            elif node_type is Data:
                child_node = self.__node__.get_data(node_identifier)
                if child_node is not None:
                    return NodeWrapper(child_node)
                else:
                    return NodeWrapper(self.__node__.add_data(node_identifier))
            else:
                raise TypeError
        elif isinstance(self.__node__, Data):
            raise LookupError
        else:
            raise TypeError

    def __add__(self, file: Union[str, Path]) -> Data:
        if isinstance(self.__node__, Data):
            if isinstance(file, str):
                file = Path(file)
            if file.exists():
                self.__node__.import_file(file, allow_overwrite=False, confirm=False)
            else:
                raise FileNotFoundError
            return self.__node__
        else:
            raise TypeError

    def __iadd__(self, file: Union[str, Path]) -> 'NodeWrapper':
        return NodeWrapper(self.__add__(file))


def walk(node: Union[Database, Bucket, DataSet, Data]):
    return NodeWrapper(node)


def __convert_to_type__(type_or_name: Union[str, type]) -> type:
    if isinstance(type_or_name, type):
        return type
    elif isinstance(type_or_name, str):
        type_or_name = type_or_name.lower()
        if type_or_name in NODE_TYPE:
            return NODE_TYPE[type_or_name]
        else:
            raise KeyError
    else:
        raise TypeError


def walk_path(node: Union[Database, Bucket, DataSet, Data],
              path: List[Union[str, Tuple[Union[str, type], str], Tuple[Union[str, type], str, bool]]]) \
        -> Union[Bucket, DataSet, Data]:
    wrapper = walk(node)
    if len(path) == 0:
        return wrapper.node
    item = path[0]
    if isinstance(item, str):
        wrapper = wrapper << item
    elif isinstance(item, tuple):
        if len(item) == 2:
            node_type, node_name = item
            required = False
        elif len(item) == 3:
            node_type, node_name, required = item
        else:
            raise AssertionError
        node_type = __convert_to_type__(node_type)
        if required:
            wrapper = wrapper >> (node_type, node_name)
        else:
            wrapper = wrapper << (node_type, node_name)
    else:
        raise TypeError
    return walk_path(wrapper.node, path[1:])



from .low.node import Root
from .bucket import Bucket
from typing import Set


class Database(Root):
    def __init__(self, name: str, path: str, version: str = 'alpha1'):
        self.__db_name__ = name
        self.__db_version__ = version
        self.__buckets__ = set()
        super().__init__(path)
        self.init_storage()
        self.init_buckets()

    def init_buckets(self):
        children = filter(lambda child: child.is_dir(), self.path.iterdir())
        for bucket in children:
            self.__buckets__.add(Bucket(bucket.name, self))

    @property
    def name(self) -> str:
        return self.__db_name__

    def rename(self, new_name: str):
        self.__db_name__ = new_name

    @property
    def version(self) -> str:
        return self.__db_version__

    @property
    def buckets(self) -> Set[Bucket]:
        return set(filter(lambda bucket: not bucket.deleted, self.__buckets__))

    @property
    def trash(self) -> Set[Bucket]:
        return set(filter(lambda bucket: bucket.deleted, self.__buckets__))

    @property
    def all_buckets(self) -> Set[Bucket]:
        return self.__buckets__

    def add_bucket(self, name: str) -> Bucket:
        if self.get_bucket(name, include_deleted=True) is not None:
            raise FileExistsError
        new_bucket = Bucket(bucket_name=name, parent=self)
        self.__buckets__.add(new_bucket)
        return new_bucket

    def insert_bucket(self, bucket: Bucket) -> Bucket:
        if bucket in self.__buckets__ \
                or self.get_bucket(bucket.name, include_deleted=True) is not None \
                or self.get_bucket(bucket.uuid, include_deleted=True):
            raise FileExistsError
        if bucket.parent is not self:
            raise AssertionError
        self.__buckets__.add(bucket)
        return bucket

    def get_bucket(self, name_or_uuid: str, include_deleted: bool = False) -> Bucket:
        target = None
        search_list = self.all_buckets if include_deleted else self.buckets
        for bucket in search_list:
            if bucket.name == name_or_uuid or bucket.uuid == name_or_uuid:
                target = bucket
        return target

    def touch_bucket(self, name: str) -> Bucket:
        target = self.get_bucket(name, include_deleted=True)
        if target is None:
            target = self.add_bucket(name)
        if target.deleted:
            raise AssertionError
        return target

    def clear_trash(self, conform: bool = True, feedback: bool = False):
        if conform and not feedback:
            return
        for bucket in self.trash:
            bucket.purge_storage(conform, feedback)
        for bucket in self.buckets:
            bucket.clear_trash(conform, feedback)
        self.__buckets__ = set(self.buckets)

    def init_storage(self):
        if not self.path.exists():
            super().init_storage()



from datetime import datetime
from typing import Union


class BackupProfile:
    def __init__(self,
                 db_name: str,
                 profile_name: Union[None, str] = None,
                 time: Union[None, datetime] = None,
                 **kwargs):
        self.__db_name__ = db_name
        self.__profile_name__ = profile_name
        self.__time__ = time
        self.parse_name()

    def parse_name(self):
        if self.__profile_name__ is not None:
            time_str = self.__profile_name__.replace(f'db_backup_{self.__db_name__}_', '').replace('.json', '')
            self.__time__ = datetime.strptime(time_str, '%Y%m%d-%H%M%S-%f')
        elif self.__time__ is not None:
            self.__profile_name__ = f"db_backup_{self.__db_name__}_{self.__time__.strftime('%Y%m%d-%H%M%S-%f')}.json"
        else:
            raise AssertionError

    @property
    def db_name(self) -> str:
        return self.__db_name__

    def rename_db(self, new_db_name: str):
        self.__db_name__ = new_db_name
        self.__profile_name__ = self.name

    @property
    def name(self) -> str:
        return f"db_backup_{self.__db_name__}_{self.__time__.strftime('%Y%m%d-%H%M%S-%f')}.json"

    @property
    def time(self) -> datetime:
        return self.__time__

    def __str__(self) -> str:
        return f"GenericBackupProfile(db='{self.__db_name__}', time='{self.time}')"

    def __repr__(self) -> str:
        return f"GenericBackupProfile(db='{self.__db_name__}', time='{self.time}')"



from typing import Tuple, Union


class BackupBackend:
    def __init__(self, **kwargs):
        self.__is_connected__ = False

    def ping(self) -> Union[bool, Tuple[bool, float]]:
        raise NotImplementedError

    def connect(self):
        raise NotImplementedError

    @property
    def is_connected(self) -> bool:
        return self.__is_connected__

    def create_backup(self, **kwargs):
        raise NotImplementedError

    def sync_backup(self, **kwargs):
        raise NotImplementedError

    def list_backups(self, **kwargs):
        raise NotImplementedError

    def fetch_backup(self, **kwargs):
        raise NotImplementedError

    def recover_from_backup(self, **kwargs):
        raise NotImplementedError
